#include <stdio.h>
#include "ports.h"
#include "cprintf.h"

/* LEDS are:

   LED0 = P8.3
   LED1 = P8.4
   LED2 = P3.3
   LED3 = P3.4

   TxD0	= P0.4
   TxD1	= P2.6
*/

static void dump_regs (void *frame, void *pc);

volatile int timer = 0;

//extern tpu0_handler ();

static void __attribute__((interrupt))
tpu0_handler ()
{
  timer ++;
  static int old_timer = 0;
  int new_timer;

#if 1
  new_timer = timer & 0xff00;
  if (new_timer != old_timer) /* about 4x per second */
    {
      old_timer = new_timer;
      asm("setpsw I");

      static int foo = 0;
      foo = ~foo;

      DR[8] |= 0x18;
      if (foo & 1)
	DR[8] &= ~0x08;
      else
	DR[8] &= ~0x10;

      dump_regs (__builtin_frame_address (0), __builtin_return_address (0));
    }
#endif
}

/* LCD is:

   14 D7  DLCDD7  P9.7  pass MSN first, then LSN.
   13 D6  DLCDD6  P9.6
   12 D5  DLCDD5  P9.5
   11 D4  DLCDD4  P9.4
   10 -
    9 -
    8 -
    7 -
    6 En  DLCDE   P8.6  Pulse H
    5 RW  Vss
    4 RS  DLCDRS  P8.5
    3 Vo  -
    2 Vdd Vdd
    1 Vss Vss

    E RS Func
    0 -  disabled
    1 0  instruction write
    1 1  data write

     0  1  2  3  4  5  6  7  addresses
    40 41 42 43 44 45 46 47
*/

static inline void
wait_ns_1 (int n)
{
  /* This takes 1 cycle for the add, and 3 cycles for the branch.  At
     100 MHz, that's 40 nS per iteration.  */
  __asm__ __volatile__ ("\n\
1:	add	#-1, %0\n\
	bgt	1b"
	 : "=r" (n) : "0" (n));
}

/* Scale each of these to the number of iterations of the above loop,
   rounding up - delays are minimums, not maximums.  */
#define wait_ns(x) wait_ns_1 ((x + 39) / 40)
#define wait_us(x) wait_ns_1 (((x*1000) + 39) / 40)
#define wait_ms(x) wait_ns_1 (((x*1000000) + 39) / 40)

static inline void
lcd_nibble (int n)
{
  DR[9] = n;
  DR[8] |=  0x40;
  wait_ns (450);
  DR[8] &= ~0x40;
  wait_ns (450);
}

static void
lcd_command (int c)
{
  DR[8] &= ~0x20;
  lcd_nibble (c);
  lcd_nibble (c << 4);
}

static void
lcd_data (int d)
{
  DR[8] |=  0x20;
  lcd_nibble (d);
  lcd_nibble (d << 4);
  wait_us (40);
}

static void
lcd_init ()
{
  DDR[8] |= 0x60;
  DDR[9] |= 0xf0;
  DR[8] = 0; /* disable */
  wait_ms (31);

  DR[8] &= ~0x20;

  /* We don't know the state of the LCD.  There are three states we
     could be in:

     1. 8-bit mode.  We issue three "8-bit" commands, then a 4-bit
       command.

     2. 4-bit mode, in sync.  We issue two 8-bit commands, then a
        4-bit command.

     3. 4-bit mode, between nibbles.  We issue garbage followed by an
        8-bit command, followed by a 4-bit command.
  */
  /*                    case 1	case 2	case 3 */
  lcd_nibble (0x30); /* 8:8bit	4:8bitH	4:garbage */
  wait_us (40);
  lcd_nibble (0x30); /* 8:8bit	4:8bitL	4:8bitH */
  wait_us (40);
  lcd_nibble (0x30); /* 8:8bit	8:8bit	4:8bitL */
  wait_us (40);
  /* We now know what state the LCD is in - we're in 8-bit mode.  */
  lcd_nibble (0x20); /* Switch to 4-bit interface, from 8-bit mode.  */
  wait_us (40);
  /* We now know we're in 4-bit mode.  */
  lcd_command (0x2c); /* LCD on, two rows */
  wait_us (40);

  lcd_command (0x0e); /* display on, cursor on, blink off */
  wait_ms (4);

  lcd_command (0x06); /* increment, shift off */
  wait_us (40);
  lcd_command (0x01); /* clear screen */
  wait_ms (4);
}

lcd_putstr (int row, int col, char *str)
{
  lcd_command (0x80 + row * 0x40 + col);
  wait_us (39);
  while (*str)
    lcd_data (*str++);
}

static void
hexify (char *ocp, unsigned long v, int r)
{
  static char hex[] = "0123456789ABCDEF";
  char buf[9];
  int i, c;
  char *cp = buf;

  cp += 8;
  *cp = 0;
  for (i=0; i<8; i++)
    {
      *--cp = hex[v & 0x0f];
      v >>= 4;
    }
  for (c=0; c<8 && *cp == ocp[c]; c++)
    cp++;
  lcd_putstr (r, c, cp);
  for (c=0; c<8; c++)
    ocp[c] = *cp++;
}

static void
dump_regs (void *frame, void *pc)
{
  static char pbuf[9] = "xxxxxxxx";
  static char sbuf[9] = "xxxxxxxx";

  hexify (pbuf, (unsigned long)frame, 0);
  hexify (sbuf, (unsigned long)pc, 1);
}

main()
{
  int i, j;

  setbuf(stdout, NULL);

  /* Power up the peripherals.  */
  MSTPCRA = 0;
  MSTPCRB = 0;
  MSTPCRC = 0;

  DR[8] = 0x18;
  DR[3] = 0x18;

  DDR[3] = 0x18;
  DDR[8] = 0x18;

  DDR[11] = 0xff;
  DDR[12] = 0xff;

  /* Timers */
  _vectors[IVEC_TPU0_TGI0A] = tpu0_handler;
  IER0D |= 0x01;
  IPR4C = 5;

  /* Timer 0 is our millisecond interrupt timer.  */
  TPU[0].tcr = 0x20; /* TGRA, PCLK = 50 MHz */
  TPU[0].tier = 0x01; /* interrupt on TGRA */
  TPU[0].tgra = 50000; /* should be 1 mS */
  TSTRA |= 0x01; /* start timer */

  set_system_clock_speed ();
  setup_interrupts ();

  tty_init ();

  lcd_init ();

  lcd_putstr (0, 0, "DJ      ");
  lcd_putstr (1, 0, "Delorie ");

  while (1)
    {
      static int t;

      t = (t+1) % 4;
      tty_putc("-\\|/"[t]);
      tty_putc(8);
    }

  tty_putc(3); tty_putc(3); tty_putc(0);
}
