#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "ogg/ogg.h"
#include "vorbis/codec.h"
#include "vorbis/vorbisenc.h"

#define MAXSIZE 512
#include <stdint.h>
#include "iodefine.h"
#include "ivects.h"
#include "tty.h"
#include "cprintf.h"
#include "timer.h"
#include "sram.h"
#include "critsec.h"
#include "usb.h"

/* If you change the rate, change libvorbis/lib/vorbisenc.c also  */
#define RECORD_RATE	14000
#define RECORD_QUALITY	1.0

#define SAMPLE_BUFSZ	512
#define RECORD_TIME_MS	20000

#define BUFFER_SIZE	8192

#define LED0	PORT2.DR.BIT.B5
#define LED1	PORT2.DR.BIT.B4
#define LED2	PORT2.DR.BIT.B3
#define LED3	PORT2.DR.BIT.B2
#define LED4	PORT2.DR.BIT.B1

unsigned long samples_saved = 0;
unsigned long samples_lost = 0;

/*--------------------------------------------------------------*/

/* PWM out P8.2 (MTIOC4A-B) */

static void
pwm_init ()
{
  MSTP(MTU4) = 0;

  IOPORT.PFCMTU.BIT.MTUS4 = 0;
  MTUA.TOER.BIT.OE4A = 1;

  MTU4.TCR.BYTE = 0x40; /* clear by TGRB match, PCLK/1 = 48 MHz */
  MTU4.TIORH.BYTE = 0x21;
  MTU4.TIORL.BYTE = 0x21;
  MTU4.TGRA = SAMPLE_BUFSZ/4;
  MTU4.TGRB = SAMPLE_BUFSZ * 2;
  MTU4.TMDR.BYTE = 0x02; /* PWM mode 1 */
  MTUA.TSTR.BIT.CST4 = 1; /* start timer */
}

static void inline
pwm_value (int val)
{
  MTU4.TGRA = val;
}

/*--------------------------------------------------------------*/

static char * sbase = 0;

void *
sbrk (int adj)
{
  char *prev_sbase;
  if (sbase == 0)
    sbase = sram_base ();
  prev_sbase = sbase;
  sbase += adj;
  /*cprintf("sbrk(%l) = %l..%l\n", adj, prev_sbase, sbase);*/
  return prev_sbase;
}

/*--------------------------------------------------------------*/
/* Generic Buffer Functions					*/

typedef struct {
  uint16_t left;
  uint16_t right;
} BufferEntry[BUFFER_SIZE];

typedef struct {
  int get_ptr, put_ptr;
  volatile int count;
  volatile BufferEntry data;
} Buffer;

static inline void
buffer_put (Buffer *b, uint16_t left, uint16_t right)
{
  int psw;
  int i = b->put_ptr;
  b->put_ptr = (b->put_ptr + 1) % BUFFER_SIZE;
  ENTER_CRITICAL_SECTION (psw);
  b->count ++;
  PORTC.DR.BIT.B0 = b->count >= BUFFER_SIZE-1 ? 1 : 0;
  PORTC.DR.BIT.B1 = b->count <= 1 ? 1 : 0;
  LEAVE_CRITICAL_SECTION (psw);
  b->data[i].left = left;
  b->data[i].right = right;
}

static inline void
buffer_get (Buffer *b, uint16_t *left, uint16_t *right)
{
  int psw;
  int i = b->get_ptr;
  b->get_ptr = (b->get_ptr + 1) % BUFFER_SIZE;
  ENTER_CRITICAL_SECTION (psw);
  b->count --;
  PORTC.DR.BIT.B0 = b->count >= BUFFER_SIZE-1 ? 1 : 0;
  PORTC.DR.BIT.B1 = b->count <= 1 ? 1 : 0;
  LEAVE_CRITICAL_SECTION (psw);
  *left = b->data[i].left;
  *right = b->data[i].right;
}

static inline int
buffer_full (Buffer *b)
{
  return b->count >= BUFFER_SIZE;
}

static inline int
buffer_empty (Buffer *b)
{
  return b->count <= 0;
}

Buffer adc_buffer;

/*--------------------------------------------------------------*/
/* ADC Functions						*/

static int p6 = 0;

static void __attribute__((interrupt))
adc_handler (void)
{
  uint16_t left, right;
  left  = S12AD.ADDRA;
  right = S12AD.ADDRB;
  if (!buffer_full (&adc_buffer))
    {
      buffer_put (&adc_buffer, left, right);
      samples_saved ++;
    }
  else
    {
      PORT7.DR.BIT.B5 = 1;
      samples_lost ++;
    }
  pwm_value (adc_buffer.count);
}

void
adc_init (void)
{
  PORT4.DDR.BYTE = 0;
  PORT4.ICR.BYTE = ~0;
  MSTP(S12AD) = 0;
  S12AD.ADANS.WORD = 0x0003; /* scan channels 0 and 1 */
  S12AD.ADADS.WORD = 0x0003;
  S12AD.ADADC.BYTE = 3;
  S12AD.ADCSR.BYTE = 0x1e;
  S12AD.ADCER.BIT.ADRFMT = 1; /* left aligned data */
  S12AD.ADSTRGR.BYTE = 0x3; /* TRGAN_0 */

  _vectors[VECT(S12AD,ADI)] = adc_handler;
  IEN(S12AD,ADI) = 1;
  IPR(S12AD,ADI) = 5;

  MSTP(MTU0) = 0;
  MTU1.TCR.BYTE = 0x20; /* TGRA, PCLK/1 = 48 MHz */
  MTU1.TIER.BIT.TTGE = 1; /* ADC trigger enabled */
  MTU1.TGRA = 1088; /* should be 44100 KHz */
  MTUA.TSTR.BIT.CST1 = 0; /* stop timer */
}

void
adc_start (void)
{
  MTUA.TSTR.BIT.CST1 = 1; /* start timer */
}

void
adc_stop (void)
{
  MTUA.TSTR.BIT.CST1 = 0; /* stop timer */
}

void
adc_rate (int samples_per_second)
{
  MTU1.TGRA = 48000000 / samples_per_second;
}

/*--------------------------------------------------------------*/

void
ogg_to_usb (ogg_stream_state *os)
{
  ogg_page op;

  while (ogg_stream_flush (os, &op))
    {
      usb_write (op.header, op.header_len);
      usb_write (op.body, op.body_len);
    }
}

void
ogg_maybe_to_usb (ogg_stream_state *os)
{
  ogg_page op;

  while (ogg_stream_pageout (os, &op))
    {
      usb_write (op.header, op.header_len);
      usb_write (op.body, op.body_len);
    }
}

int
main(void)
{
  uint16_t min, max;
  unsigned long last_print = 0;
  unsigned long start_time_ms;
  char *sb;
  unsigned long sl;
  unsigned long *sla;
  unsigned long *slb;
  int i;

  /*                     IPB P   */
  SYSTEM.SCKCR.LONG = 0x00010100;

  PORT5.DDR.BIT.B3 = 1;

  PORT2.DDR.BYTE = 0xff;
  PORT2.DR.BYTE = 0x00;

  PORT8.DDR.BYTE = 0xff;
  PORT8.DR.BYTE = 0;
  PORT7.DDR.BYTE |= 0xff;
  PORT7.DR.BYTE = 0;

  setbuf (stdout, NULL);
  setbuf (stderr, NULL);

  sram_init ();

  timer_init ();

  pwm_init ();

  tty_init ();
  tty_puts("\nOGG RECORDER\n");

  PORTC.DDR.BYTE = ~0;

  printf("printf works.\n");

  sb = sram_base ();
  sl = sram_size ();
  printf("sram: %d kB : 0x%08x..%08x\n", sl/1024, sb, sb + sl - 1);

  tty_puts ("hit a key...");

#ifdef TEST_SRAM_TIMINGS
  sla = (unsigned long *)sb;
  slb = (unsigned long *)(sb + 0x44);
  *sla = 0;
  *slb = ~0;
  asm("# DJ");
  while (1)
    {
      unsigned long a, b;

      asm("# 1");
      a = *sla;
      b = *slb;
      asm("# 2" :: "r" (a), "r" (b));
      *sla = b;
      *slb = a;
      asm("# 3" : "=r" (sla), "=r" (slb) : "0" (sla), "1" (slb));
    }
#endif

  while (tty_getc () == 3)
    ;
  tty_puts("\n");

  usb_init ();
  while (!usb_connected)
    ;
  tty_puts("\n");


  adc_init ();
  adc_rate (RECORD_RATE);

  char *buf;
  int r, n;
  int done;

  ogg_stream_state os;
  ogg_page og;
  ogg_packet op;
  int saw_vorbis = 0;

  vorbis_info vi;
  vorbis_comment vc;
  vorbis_dsp_state v;
  vorbis_block vb;

  float **pcm;


  vorbis_info_init (&vi);
  r = vorbis_encode_init_vbr (&vi, 2, RECORD_RATE, RECORD_QUALITY);
  if (r)
    cprintf ("encoder not found: %d\n", r);
  vorbis_comment_init (&vc);
  vorbis_comment_add_tag(&vc,"ENCODER","DJ's ogg-record demo");

  vorbis_analysis_init (&v, &vi);
  vorbis_block_init (&v, &vb);

  ogg_stream_init (&os, 12345);

  ogg_packet header;
  ogg_packet header_comm;
  ogg_packet header_code;

  vorbis_analysis_headerout (&v, &vc, &header, &header_comm, &header_code);
  ogg_stream_packetin (&os, &header);
  ogg_stream_packetin (&os, &header_comm);
  ogg_stream_packetin (&os, &header_code);

  ogg_to_usb (&os);

  adc_start ();
  start_time_ms = timer_ms;

  done = 0;
  while (! done)
    {
      done = (timer_ms - start_time_ms) > RECORD_TIME_MS;

      if (done)
	{
	  adc_stop ();
	  vorbis_analysis_wrote (&v, 0);
	}
      else
	{
	  int i;
#ifdef MINMAX
	  float max, min;
#endif
	  float **buffer=vorbis_analysis_buffer (&v, SAMPLE_BUFSZ);


#ifdef MINMAX
	  max = -32768.0;
	  min = 32768.0;
#endif
	  PORT8.DR.BIT.B0 = 1;
	  for (i=0; i<SAMPLE_BUFSZ; i++)
	    {
	      uint16_t left, right;
	      if (buffer_empty (&adc_buffer))
		{
		  PORT7.DR.BIT.B5 = 0;
		  while (buffer_empty (&adc_buffer))
		    asm("");
		}
	      buffer_get (&adc_buffer, &left, &right);
	      pwm_value (adc_buffer.count);
	      buffer[0][i] = (left  - 32768.0) / 32768.0;
	      buffer[1][i] = (right - 32768.0) / 32768.0;
#ifdef MINMAX
	      if (max < buffer[0][i])
		max = buffer[0][i];
	      if (min > buffer[0][i])
		min = buffer[0][i];
#endif
	    }
	  PORT8.DR.BIT.B0 = 0;
#ifdef MINMAX
	  printf("min %f max %f\n", min, max);
#endif
	  vorbis_analysis_wrote (&v, SAMPLE_BUFSZ);

	}

      while (vorbis_analysis_blockout (&v, &vb) == 1)
	{
#if 1
      PORT7.DR.BIT.B6 = 1;
	  vorbis_analysis(&vb,NULL);
      PORT7.DR.BIT.B6 = 0;
	  vorbis_bitrate_addblock(&vb);

	  while(vorbis_bitrate_flushpacket (&v, &op))
	    {
	      ogg_stream_packetin(&os,&op);
	      ogg_maybe_to_usb (&os);
	    }
#else
      PORT7.DR.BIT.B6 = 1;
	  vorbis_analysis(&vb,&op);
      PORT7.DR.BIT.B6 = 0;
	  ogg_stream_packetin(&os,&op);
	  ogg_maybe_to_usb (&os);
#endif
	}
    }
  ogg_to_usb (&os);

  ogg_stream_clear(&os);
  vorbis_block_clear(&vb);
  vorbis_dsp_clear(&v);
  vorbis_comment_clear(&vc);
  vorbis_info_clear(&vi);

  printf("done!\n");
  printf("rate: %d, quality: %4.2f, bufs: %d/%d\n",
	 RECORD_RATE, RECORD_QUALITY, SAMPLE_BUFSZ, BUFFER_SIZE);
  printf("samples saved: %ld, lost: %ld (%d%%)\n",
	 samples_saved, samples_lost,
	 samples_lost * 100 / (samples_lost + samples_saved));
  usb_uninit ();
  tty_putc(3);
  tty_putc(3);
  tty_putc(0);
  for (;;);
}
